/*
 * faulty.c -- a module which generates an oops when read
 * $Id: faulty.c,v 1.3 2000/05/29 20:41:25 rubini Exp $
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/module.h>

#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/types.h>  /* size_t */

#include "sysdep.h" /* linux 2.0, 2.2, 2.4 compatibility macros */

int faulty_major=0;

char faulty_buf[1024];

ssize_t faulty_read (struct file *filp, char *buf, size_t count, loff_t *pos)
{
    int ret;
    printk(KERN_DEBUG "read: file %p, buf %p, count %li\n", filp, buf,
	   (long)count);
    /*
     * this line oopses with 2.0, but doesn't oops with 2.2 and later,
     * because of the new exception handling mechanism
     */
    ret = copy_to_user(buf, faulty_buf, count);
    if (ret < 0) return ret; /* 2.2 and 2.4 */
    return count; /* never reached, actually */
}

ssize_t faulty_write (struct file *filp, const char *buf, size_t count,
		loff_t *pos)
{
    /* make a real fault by dereferencing a NULL pointer */
    *(int *)0 = 0;
    return 0;
}

#ifdef LINUX_20 /* wrappers for 2.0 */
int faulty_read_20 (struct inode *ino, struct file *f, char *buf, int count)
{
    return (int) faulty_read(f, buf, count, &f->f_pos);
}

int faulty_write_20 (struct inode *ino, struct file *f, const char *b, int c)
{
    return (int) faulty_write(f, b, c, &f->f_pos);
}


#define faulty_read faulty_read_20
#define faulty_write faulty_write_20
#endif  /* LINUX_20 */


struct file_operations faulty_fops = {
    read:  faulty_read,
    write: faulty_write,
};


int init_module(void)
{
    int result;

    /*
     * Register your major, and accept a dynamic number
     */
    result = register_chrdev(faulty_major, "faulty", &faulty_fops);
    if (result < 0) return result;
    if (faulty_major == 0) faulty_major = result; /* dynamic */

    return 0;
}

void cleanup_module(void)
{
    unregister_chrdev(faulty_major, "faulty");
}



